#lang scheme

(require redex/reduction-semantics
         "source.rkt"
         "target.rkt")

(provide compile)

;; the compilation function defined in the paper, section 3.3

(define-metafunction sourcelang
  compile : prog -> any
  [(compile (s ...)) (tsort ((free chan) ... (compile-state s) ...))
                     (where ((chan_1 ...) ...) ((chans-of s) ...))
                     (where (chan ...) (no-dups/ch (chan_1 ... ...)))])


(define-metafunction sourcelang
  compile-state : s -> (any ...)
  [(compile-state a) a]
  [(compile-state (select a ...)) (nu (any ...) ((active any a) ... (open ((any a) ...))))
                                  (where (any ...) ,(variables-not-in `() (map (lambda (dc) 'v) (term (a ...)))))])


(define-metafunction sourcelang
  chans-of : s -> (chan ...)
  [(chans-of a) ((chan-of a))]
  [(chans-of (select a ...)) (no-dups/ch ((chan-of a) ...))])

(define-metafunction sourcelang
  chan-of : a -> chan
  [(chan-of (in chan)) chan]
  [(chan-of (out chan)) chan])

;; no-dups/ch : take a list of chans, return the list without duplications
(define-metafunction sourcelang
  no-dups/ch : (chan ...) -> (chan ...)
  [(no-dups/ch ()) ()]
  [(no-dups/ch (chan chan_0 ... chan chan_1 ...)) (no-dups/ch (chan chan_0 ... chan_1 ...))]
  [(no-dups/ch (chan chan_0 ...)) (chan chan_1 ...)
                                  (where (chan_1 ...) (no-dups/ch (chan_0 ...)))])

(test-equal (term (chan-of (in ch1))) (term ch1))
(test-equal (term (chan-of (out ch2))) (term ch2))
(test-equal (term (chans-of (in ch1))) (term (ch1)))
(test-equal (term (chans-of (out ch2))) (term (ch2)))
(test-equal (term (chans-of (select (in ch1) (out ch2) (in ch3)))) (term (ch1 ch2 ch3)))
(test-equal (term (chans-of (select (in ch1) (out ch2) (in ch1)))) (term (ch1 ch2)))


(test-equal (term (compile-state (out ch1)))
            (term (out ch1)))
(test-equal (term (compile-state (select (out ch1))))
            (term (nu (v) ((active v (out ch1)) (open ((v (out ch1))))))))

(test-equal (term (compile ()))
            (term (nu () ())))
(test-equal (term (compile ((out ch1))))
            (term (nu () ((out ch1) (free ch1)))))
(test-equal (term (compile ((select (out ch1)))))
            (term (nu (v) ((active v (out ch1)) (free ch1) (open ((v (out ch1))))))))
#;(test-equal (term (compile ((select (in ch1) (out ch2) (in ch1)) (select (out ch1) (in ch2))))))