#lang racket
(require "main.rkt" "util.rkt" redex)

(struct ip (p w))

(define interesting-programs
  (list (ip 1 1)
        (ip (term ((fn x x) 1)) 1)
        (ip (term (let x 2 x)) 2)
        (ip (term (let x ((fn x x) 1) x)) 1)
        (ip (term (let x 1 (let y (fn z x) (y 2)))) 1)
        (ip (term (let x 3 (let y (fn z (let w 1 x)) (let a (y 4) a)))) 3)
        (ip (term (let f (fn x x) (let a1 (f 1) (let a2 (f a1) (f a2))))) 1)
        (ip (term (let a1 ((fn x x) 12) (let f (fn x x) (let a2 (fn y (f a1)) (a2 11))))) 12)
        (ip (term (let g (fn f (let q (f 1) (q 2))) (g (fn x (fn y x))))) 1)
        
        ;; the analysis should produce both 1 and 2
        (ip (term (let g (fn x x) (let a (g 1) (let b (g 2) b)))) 2)

        ;; these programs should do poorly in the analysis
        (ip (term (let g (fn z (let f (fn y 1) (let x 2 (f x))))
                    (let f (fn w 3) (let x (g 1) (f x)))))
            3)
        (ip (term (let g (fn z (let f (fn y 1) (let x 2 (f x))))
                    (let q (g 5)
                      (let f (fn w 3) (let x 4 (f x))))))
            3)))

;; lookup
(test-equal (term (lookup z (• z 1)))
            (term 1))
(test-equal (term (lookup z ((• z 2) q 1)))
            (term 2))
(test-equal (term (lookup w ((• z 2) q 1)))
            (term #f))

;; μ
(test-equal (term (μ 1 •)) (term 1))
(test-equal (term (μ z (• z 2))) (term 2))
(test-equal (term (μ (fn x 1) •)) (term (close (fn x 1) •)))

;; fv 
(test-equal (term (fv x)) (term (x)))
(test-equal (term (fv (fn x y))) (term (y)))
(test-equal (term (fv 1)) (term ()))
(test-equal (term (fv (y x))) (term (x y)))
(test-equal (term (fv (x x))) (term (x)))
(test-equal (term (fv (let x y (x y)))) (term (y)))
(test-equal (term (fv (let x (y z) x))) (term (y z)))
(test-equal (term (fv (let x (y x) x))) (term (x y)))

;; bind-fv
(test-equal (bind-fv '(x y))
            `(let x (fn fvx 100) (let y (fn fvx 111) (x y))))

;; CaEK
(test--> CaEK
         (term (1 • (z 2 • stop)))
         (term (2 (• z 1) stop)))
(test--> CaEK
         (term (1 (• p 3) (z 2 (• q 4) stop)))
         (term (2 ((• q 4) z 1) stop)))
(test--> CaEK
         (term ((fn x p) (• p 3) (z 2 (• q 4) stop)))
         (term (2 ((• q 4) z (close (fn x p) (• p 3))) stop)))

(test--> CaEK
         (term ((let x (fn z y) (x p)) ((• p 3) y 2) stop))
         (term ((x p) (((• p 3) y 2) x (close (fn z y) ((• p 3) y 2))) stop)))

(test--> CaEK
         (term (((fn x x) (fn y y)) • stop))
         (term (x (• x (close (fn y y) •)) stop)))

(test--> CaEK
         (term ((let z ((fn x x) (fn y y)) z) • stop))
         (term (x (• x (close (fn y y) •)) (z z • stop))))

(test-->> CaEK
          (term ((let x 1 x) • (xr xr • stop)))
          (term (xr (• xr 1) stop)))

(test-->> CaEK
          (term ((let f (fn x x) (let a1 (f 1) (let a2 (f a1) (f a2)))) • (xr xr • stop)))
          (term (xr (• xr 1) stop)))

(test-->> CaEK
          (term ((let a1 ((fn x x) 12) (let f (fn x x) (let a2 (fn y (f a1)) (a2 11)))) • (xr xr • stop)))
          (term (xr (• xr 12) stop)))

(test-->> CaEK
          (term ((let g (fn f (let q (f 1) (q 2))) (g (fn x (fn y x))))
                 •
                 (xr xr • stop)))
          (term (xr (• xr 1) stop)))

;; μc
(test-equal (term (μc 1 ()))
            (term (1)))
(test-equal (term (μc x (•)))
            (term ()))
(test-equal (term (μc x ((• x 1))))
            (term (1)))
(test-equal (term (μc x ((• x 2) •)))
            (term (2)))
(test-equal (term (μc x ((• x 1) (• x 2) (• x 3))))
            (term (1 2 3)))

;; Fc
(test-equal ((Fc '11) (list '(1 • (q 2 • stop))))
            (append (F-init 11) (list '(2 (• q 1) stop))))
(test-equal ((Fc '11) (list '((let r 1  r) • stop)))
            (append (F-init 11) (list '(r (• r 1) stop))))
(test-equal ((Fc '11) (list '(((fn q q) 1) • stop)))
            (append (F-init 11) (list '(q (• q 1) stop))))
(test-equal ((Fc '11)
             (list '((let a1 (f 1) (let a2 (f a1) (f a2)))
                     (• f (close (fn x x) •))
                     (xr xr • stop))))
            (append (F-init 11) 
                    (list '(x
                            (• x 1)
                            (a1 (let a2 (f a1) (f a2)) 
                                (• f (close (fn x x) •))
                                (xr xr • stop))))))

;; Eval        
(for ([an-ip (in-list interesting-programs)])
  (test-equal (Eval (ip-p an-ip))
              (ip-w an-ip)))

;; F vs Fc
(for ([an-ip (in-list interesting-programs)])
  (test-equal ((F (ip-p an-ip)) '())
              ((Fc (ip-p an-ip)) '())))

(for ([an-ip (in-list interesting-programs)])
  (test-equal (F-answers (lfp (F (ip-p an-ip)) '()))
              (list (ip-w an-ip)))
  (test-equal (F-answers (lfp (Fc (ip-p an-ip)) '()))
              (list (ip-w an-ip))))

;; αx
(test-equal (term (αx ((1 • stop) (2 • stop))))
            (term ((1 2) ((1 stop) (2 stop)) (•))))
(test-equal (term (αx (((fn y x) (• x 2) stop) ((fn y x) (• x 3) stop))))
            (term (((fn y x)) (((fn y x) stop)) ((• x 2) (• x 3)))))
(test-equal (term (αx (((fn y z) (• z 2) stop) ((fn y x) (• x 3) stop))))
            (term (((fn y x) (fn y z)) (((fn y x) stop) ((fn y z) stop)) ((• x 3) (• z 2)))))

;; γx
(test-equal (term (γx ((1 2) ((1 stop) (2 stop)) (•))))
            (term ((1 • stop) (2 • stop))))
(test-equal (term (γx (((fn y x)) (((fn y x) stop)) ((• x 2) (• x 3)))))
            (term (((fn y x) (• x 2) stop)
                   ((fn y x) (• x 3) stop))))

;; ∪x

(test-equal (∪x '((1) (2) (3)))
            '((1) (2) (3)))

(test-equal (∪x '((1) (2) (3))
                '((4) (5) (6)))
            '((1 4) (2 5) (3 6)))

(test-equal (∪x '((1 x) (2) (3))
                '((4 x) (5 6 7 8 9) ()))
            '((1 4 x) (2 5 6 7 8 9) (3)))

;; Fx
(test-equal (let ([f (Fx 1)])
              (f (f '())))
            (term ((1 xr)
                   ((1 (xr xr • stop)) (xr stop))
                   ((• xr 1) •))))
(test-equal (let* ([t (term (let x 1 x))]
                   [f (Fx t)])
              (f (Fx-init t)))
            (term (((let x 1 x) x) 
                   (((let x 1 x) (xr xr • stop)) (x (xr xr • stop)))
                   ((• x 1) •))))
(test-equal (let* ([t (term ((fn x x) 1))]
                   [f (Fx t)])
              (f (Fx-init t)))
            (term ((((fn x x) 1) x) 
                   ((((fn x x) 1) (xr xr • stop)) (x (xr xr • stop)))
                   ((• x 1) •))))
(test-equal (let* ([t (term (let y ((fn x x) 1) y))]
                   [f (Fx t)])
              (f (Fx-init t)))
            (term (((let y ((fn x x) 1) y) x) 
                   (((let y ((fn x x) 1) y) (xr xr • stop)) (x (y y • (xr xr • stop))))
                   ((• x 1) •))))

#;
;; can't do this because the environment structures keep growing so we get infinite things there
(for ([an-ip (in-list interesting-programs)])
  (printf "trying ~s\n" (ip-p an-ip))
  (test-equal (Fx-answers (lfp (Fx (ip-p an-ip)) '(() () ())))
              (list (ip-w an-ip))))
  
;; ρ

(test-equal (ρ (term ((1) () ())))
            (term ((1) () ())))
(test-equal (ρ (term ((1 2 3) () ())))
            (term ((1 2 3) () ())))
(test-equal (ρ (term ((1) ((x stop)) (•))))
            (term ((1) ((x stop)) (•))))
(test-equal (ρ (term ((1) ((x stop)) ((• y 1)))))
            (term ((1) ((x stop)) ((• y 1) •))))
(test-equal (ρ (term ((1) ((x stop)) ((• y (close (fn x z) (• z 2)))))))
            (term ((1) ((x stop)) ((• y (close (fn x z) (• z 2))) (• z 2) •))))
(test-equal (ρ (term ((1) ((x stop)) ((• y 1) (• y 2) (• z 3)))))
            (term ((1) ((x stop)) ((• y 1) (• y 2) (• z 3) •))))
(test-equal (ρ (term ((1)
                      ((z (x 2 (• y (close (fn x z) (• z 2))) stop)))
                      ((• y (close (fn x z) (• z 2)))))))
            (term ((1) 
                   ((2 stop) (z (x 2 (• y (close (fn x z) (• z 2))) stop))) 
                   ((• y (close (fn x z) (• z 2))) (• z 2) •))))
(test-equal (ρ (term ((1)
                      ((z (x 2 • (y 3 (• q 2) stop))))
                      (•))))
            (term ((1) 
                   ((2 (y 3 (• q 2) stop)) (3 stop) (z (x 2 • (y 3 (• q 2) stop)))) 
                   ((• q 2) •))))

;; Fρ
(test-equal (let ([f (Fρ 1)])
              (f (f '())))
            (ρ (term ((1 xr)
                      ((1 (xr xr • stop)) (xr stop))
                      ((• xr 1) •)))))
(test-equal (let* ([t (term (let x 1 x))]
                   [f (Fρ t)])
              (f (Fρ-init t)))
            (ρ (term (((let x 1 x) x) 
                      (((let x 1 x) (xr xr • stop)) (x (xr xr • stop)))
                      ((• x 1) •)))))
(test-equal (let* ([t (term ((fn x x) 1))]
                   [f (Fρ t)])
              (f (Fρ-init t)))
            (ρ (term ((((fn x x) 1) x) 
                      ((((fn x x) 1) (xr xr • stop)) (x (xr xr • stop)))
                      ((• x 1) •)))))
(test-equal (let* ([t (term (let y ((fn x x) 1) y))]
                   [f (Fρ t)])
              (f (Fρ-init t)))
            (ρ (term (((let y ((fn x x) 1) y) x) 
                      (((let y ((fn x x) 1) y) (xr xr • stop)) (x (y y • (xr xr • stop))))
                      ((• x 1) •)))))

;; ≡
(test-equal (term (≡ 1))
            (term 1))
(test-equal (term (≡ (f x)))
            (term (f x)))
(test-equal (term (≡ (let x 1 x)))
            (term x))
(test-equal (term (≡ (let x 1 ((fn x x) x))))
            (term ((fn x x) x)))
(test-equal (term (≡ (let x ((fn x x) x) 2)))
            (term 2))
(test-equal (term (≡ (let f (fn x x) (let y 1 (f y)))))
            (term (f y)))

;; ∪·
(test-equal (term (∪· ((x (1))) ((x (2)))))
            (term ((x (1 2)))))
(test-equal (term (∪· ((y (1))) ((x (2)))))
            (term ((x (2)) (y (1)))))
(test-equal (term (∪· ((y (1 2))) ((y (2 3)))))
            (term ((y (1 2 3)))))
(test-equal (term (∪·))
            (term ()))
(test-equal (term (∪· ((x ())))) ;; technically, this input isn't well-formed, but it is easier to let ∪· clean this up
            (term ()))

;; fapp
(test-equal (term (fapp () x))
            (term ()))
(test-equal (term (fapp ((a (1 2 3)) (b (4 5 6)) (c (7 8 9))) a))
            (term (1 2 3)))
(test-equal (term (fapp ((a (1 2 3)) (b (4 5 6)) (c (7 8 9))) b))
            (term (4 5 6)))
(test-equal (term (fapp ((a (1 2 3)) (b (4 5 6)) (c (7 8 9))) c))
            (term (7 8 9)))
(test-equal (term (fapp ((a (1 2 3)) (b (4 5 6)) (c (7 8 9))) d))
            (term ()))

;; μ#
(test-equal (term (μ# 1 ())) (term (1)))
(test-equal (term (μ# x ((x (1 2 3))))) (term (1 2 3)))
(test-equal (term (μ# (fn x y) ())) (term ((fn x y))))

;; ∪⊗
(test-equal (∪⊗ (term ((1) ((x (1)) (y (2))) ((a (3)))))
                (term ((1 2) ((x (3)) (y (2))) ((a (3)) (b (1 2))))))
            (term ((1 2) ((x (1 3)) (y (2))) ((a (3)) (b (1 2))))))
           
;; F#
(test-equal ((F# 1) (term (() () ())))
            ;; effectively a F#-init test
            (term ((1) ((1 ((xr xr))) (xr (stop))) ())))
(test-equal ((F# 1) (term ((2) ((2 (stop))) ())))
            (term ((1) ((1 ((xr xr))) (xr (stop))) ())))
(test-equal ((F# 1) (term ((2) ((2 ((xs xs)))) ())))
            (term ((1 xs) ((1 ((xr xr))) (2 ((xs xs))) (xr (stop))) ((xs (2))))))
(test-equal ((F# 1) (term (((let x 2 y)) () ())))
            (term ((1 y) ((1 ((xr xr))) (xr (stop))) ((x (2))))))
(test-equal ((F# 1) (term ((((fn x y) 2)) ((((fn x y) 2) (stop))) ())))
            (∪⊗ (term ((1) ((1 ((xr xr))) (xr (stop))) ()))
                (term ((y) ((((fn x y) 2) (stop)) (y (stop))) ((x (2)))))))
(test-equal ((F# 1) (term (((let z ((fn x y) 2) 3)) ((3 (stop))) ())))
            (∪⊗ (term ((1) ((1 ((xr xr))) (xr (stop))) ()))
                (term ((y) ((3 (stop)) (y ((z 3)))) ((x (2)))))))

(for ([an-ip (in-list interesting-programs)])
  (define got (F#-answers (lfp (F# (ip-p an-ip)) '(() () ()))))
  (test-equal (and (member (ip-w an-ip) got) #t)
              #t))

;; test infinite loop and errors:
(test-equal (F#-answers (lfp (F# (term ((fn x (x x)) (fn x (x x))))) 
                             '(() () ())))
            '())

(test-equal (F#-answers (lfp (F# (term ((fn x (let f (x x) (f x)))
                                        (fn x (let f (x x) (f x)))))) 
                             '(() () ())))
            '())

(test-equal (F#-answers (lfp (F# (term (1 2))) 
                             '(() () ())))
            '())

(test-equal (F#-answers (lfp (F# (term (let x y x)))
                             '(() () ())))
            '())

(test-results)
