#lang scheme

(require redex/reduction-semantics
         "grammar.rkt")

(define ->
  (reduction-relation
   SR
   (==> ((block (def f (x : σ) : τ = e) f) v)
        (subst (subst e x v) f (block (def f (x : σ) : τ = e) f))
        "cbv")
   (==> (o v)
        (δ o v)
        "prim")
   (==> (block v r ... e)
        (block r ... e)
        "block-next")
   (==> (block e)
        e
        "block-done")
   (==> (block (val x : τ = v) r ... e)
        (block r_^’ ... e_^’)
        (where (block r_^’ ... e_^’) (subst (block r ... e) x v))
        "val")
   (==> (block (def f (x : σ) : τ = e_1) r ... e_2)
        (subst (block r ... e_2) f v)
        (where v (block (def f (x : σ) : τ = e_1) f))
        (side-condition (term (is-def? (r ...) f e_2)))
        "def")
   
   with
   [(--> (in-hole E exp1) (in-hole E exp2))
    (==> exp1 exp2)]))

(define-metafunction SR
  [(is-def? (r ...) f e)
   ,(not (and (null? (term (r ...)))
              (eq? (term f) (term e))))])

(define ->sr
  (extend-reduction-relation
   -> SR
   (--> (in-hole E (reset (in-hole G (shift (name v (block (def f (x : ((⇒ (A ε) (B ε)) ε)) : (C ε) = e) f))))))
        (in-hole E (v (block (def y (z : (A ε)) : (B ε) = (reset (in-hole G z))) y)))
        (fresh y)
        (fresh z)
        "shift")
   (--> (in-hole E (reset v))
        (in-hole E v)
        "reset")))

(define-metafunction SR
  δ : o v -> v
  [(δ inc number) ,(add1 (term number))]
  [(δ dbl number) ,(* 2 (term number))]
  [(δ len string) ,(string-length (term string))])

(define-metafunction SR
  subst : any x e -> any
  
  [(subst (block (name r (def f (x : σ) : τ = e_2)) s ...) f e_1)
   (block r s ...)]
  [(subst (block (name r (def f (x : σ) : τ = e_2)) s ...) x e_1)
   (block r s_^’ ...)
   (where (block s_^’ ...) (subst (block s ...) x e_1))]
  [(subst (block (def f (y : σ) : τ = e_2) r ...) x e_1)
   (block
    (def f_^’ (y_^’ : σ) : τ = (subst (subst-var f f_^’ (subst e_2 y y_^’)) x e_1))
    s ...)
   (where (f_^’ y_^’) ,(variables-not-in (term (x e_1 e_2 r ...)) (term (f y))))
   (where (block s ...) (subst (subst (block r ...) f f_^’) x e_1))]
  
  [(subst (block (val x : τ = e_2) r ...) x e_1)
   (block (val x : τ = (subst e_2 x e_1)) r ...)]
  [(subst (block (val y : τ = e_2) r ...) x e_1)
   (block
    (val y_^’ : τ = (subst e_2 x e_1))
    s ...)
   (where y_^’ ,(variable-not-in (term (x e_1 r ...)) (term y)))
   (where (block s ...) (subst (subst-var y y_^’ (block r ...)) x e_1))]
  
  [(subst (block e_1 r_2 r_3 ...) x e)
   (block (subst e_1 x e) s_2 s_3 ...)
   (where (block s_2 s_3 ...) (subst (block r_2 r_3 ...) x e))]
  [(subst (block e_2) x e_1)
   (block (subst e_2 x e_1))]
  
  [(subst x x e) e]
  [(subst y x e) y]
  
  [(subst (any ...) x e)
   ((subst any x e) ...)]
  [(subst any x e) any])

(define-metafunction SR
  subst-var : x x any -> any
  [(subst-var x y x) y]
  [(subst-var x y z) z]
  [(subst-var x y (any ...))
   ((subst-var x y any) ...)]
  [(subst-var x y any) any])

(provide -> ->sr subst)
