#lang racket

(require redex)

(provide (all-defined-out))

(define-syntax-rule (redex-test-suite . body)
  (run-redex-test-suite (lambda () . body)))

(define in-redex-test-suite? (make-parameter #f))

(define (run-redex-test-suite f)
  (if (in-redex-test-suite?)
    (f)
    (parameterize ([in-redex-test-suite? #t])
      (f)
      (test-results))))

(define-language empty-lang)

(define-relation empty-lang
  [(¬ #f)])

(define-relation empty-lang
  [(≈ () ...)]
  [(≈ (any_1 any_2 ...) ...)
   (≈ (any_2 ...) ...)])

(define-relation empty-lang
  [(≡)]
  [(≡ any)]
  [(≡ any any any_other ...) (≡ any any_other ...)])

(define-relation empty-lang
  [(∧)]
  [(∧ any_1 any_2 ...) any_1 (∧ any_2 ...)])

(define-relation empty-lang
  [(∨ any_1 ... any any_2 ...) any])

(define-relation empty-lang
  [(∈ any {any_a ... any any_b ...})])

(define-metafunction empty-lang
  ∪ : {any ...} ... -> {any ...}
  [(∪) {}]
  [(∪ {any ...}) {any ...}]
  [(∪ {any_a ... any any_b ...} {any_c ... any any_d ...} {any_e ...} ...)
   (∪ {any_a ...     any_b ...} {any_c ... any any_d ...} {any_e ...} ...)]
  [(∪ {any_a ...} {any_b ...} {any_c ...} ...)
   (∪ {any_a ...   any_b ...} {any_c ...} ...)])

(define-metafunction empty-lang
  ∩ : {any ...} ... -> {any ...}
  [(∩) {}]
  [(∩ {any_a ...} ... {} {any_b ...} ...) {}]
  [(∩ {any any_a ...} {any_b ...} ...)
   (∪ {any} (∩ {any_a ...} {any_b ...} ...))
   (side-condition (term (∧ (∈ any {any_b ...}) ...)))]
  [(∩ {any any_a ...} {any_b ...} ...)
   (∩ {any_a ...} {any_b ...} ...)])

(define-metafunction empty-lang
  ∖ : {any ...} {any ...} ... -> {any ...}
  [(∖ {any ...}) {any ...}]
  [(∖ {any_a ... any any_b ...} {any_c ... any any_d ...} {any_e ...} ...)
   (∖ {any_a ...     any_b ...} {any_c ...     any_d ...} {any_e ...} ...)]
  [(∖ {any_a ...} {any_b ...} {any_c ...} ...)
   (∖ {any_a ...}             {any_c ...} ...)])

(define (test-empty)
  (redex-test-suite
  
    (test-equal (term (≈ () () ())) #t)
    (test-equal (term (≈ (a) (b) (c))) #t)
    (test-equal (term (≈ () (a) (b c))) #f)
  
    (test-equal (term (≡ (a b c) (a b c))) #t)
    (test-equal (term (≡ (a b c) (d e f))) #f)
  
    (test-equal (term (∧)) #t)
    (test-equal (term (∧ #t)) #t)
    (test-equal (term (∧ #f)) #f)
    (test-equal (term (∧ #t #t)) #t)
    (test-equal (term (∧ #t #f)) #f)
    (test-equal (term (∧ #f #t)) #f)
    (test-equal (term (∧ #f #f)) #f)
    (test-equal (term (∧ #t #t #t)) #t)
    (test-equal (term (∧ #t #f #t)) #f)
    (test-equal (term (∧ #f #t #f)) #f)
    (test-equal (term (∧ #f #f #f)) #f)
  
    (test-equal (term (∨)) #f)
    (test-equal (term (∨ #t)) #t)
    (test-equal (term (∨ #f)) #f)
    (test-equal (term (∨ #t #t)) #t)
    (test-equal (term (∨ #t #f)) #t)
    (test-equal (term (∨ #f #t)) #t)
    (test-equal (term (∨ #f #f)) #f)
  
    (test-equal (term (∈ b (a b c))) #t)
    (test-equal (term (∈ b (d e f))) #f)
  
    (test-equal (term (∪ {a b c} {b c d} {c d e})) '(a b c d e))
  
    (test-equal (term (∩ {a b c} {b c d} {c d e})) '(c))
  
    (test-equal (term (∖ (a b c) (b c d))) '(a))))
